--*****************************************************************************
--*  Copyright (c) 2014-2015 by Michael Fischer (www.emb4fun.de) 
--*  All rights reserved.
--*
--*  Redistribution and use in source and binary forms, with or without 
--*  modification, are permitted provided that the following conditions 
--*  are met:
--*  
--*  1. Redistributions of source code must retain the above copyright 
--*     notice, this list of conditions and the following disclaimer.
--*  2. Redistributions in binary form must reproduce the above copyright
--*     notice, this list of conditions and the following disclaimer in the 
--*     documentation and/or other materials provided with the distribution.
--*  3. Neither the name of the author nor the names of its contributors may 
--*     be used to endorse or promote products derived from this software 
--*     without specific prior written permission.
--*
--*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
--*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
--*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
--*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
--*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
--*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
--*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
--*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
--*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
--*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
--*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
--*  SUCH DAMAGE.
--*
--*****************************************************************************
--*  History:
--*
--*  19.06.2014  mifi  First version.
--*  23.05.2015  mifi  Use correct GPMC identifier.
--*  23.07.2015  mifi  Make it more universal.
--*****************************************************************************

--*****************************************************************************
--*  DEFINE: Library                                                          *
--*****************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


--*****************************************************************************
--*  DEFINE: Entity                                                           *
--*****************************************************************************

--
-- CPU <-> Wishbone Master
--
-- CPU_XX are the original not synchronized bus signals. 
-- The synchronisation will be done by the module here.
--
-- In case of the CPU it could be a GPMC (TI) or FSMC (STM) bus.
--
entity cpu_wb_master is
   generic (
      ADDRESS_WIDTH : integer := 11
   );
   port (  
      clk_i      : in    std_logic := '0';
      rst_i      : in    std_logic := '0';
      
      -- CPU_A is a word address
      CPU_A      : in    std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
      CPU_D      : inout std_logic_vector(15 downto 0);
      CPU_CS_n   : in    std_logic := '1';
      CPU_OE_n   : in    std_logic := '1';
      CPU_WE_n   : in    std_logic := '1';
      CPU_WAIT_n : out   std_logic;
      
      -- Wishbone Master
      wbm_stb_o  : out   std_logic;
      wbm_cyc_o  : out   std_logic;
      wbm_we_o   : out   std_logic;
      wbm_ack_i  : in    std_logic := '0';

      wbm_adr_o  : out   std_logic_vector(ADDRESS_WIDTH-1 downto 0);  
      wbm_dat_o  : out   std_logic_vector(15 downto 0);  
      wbm_dat_i  : in    std_logic_vector(15 downto 0) := (others => '0')
   );
end entity cpu_wb_master;


--*****************************************************************************
--*  DEFINE: Architecture                                                     *
--*****************************************************************************

architecture syn of cpu_wb_master is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------

   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------

   --
   -- Control signal defines
   --   
   signal cs_s1  : std_logic := '0';  
   signal cs_s2  : std_logic := '0';  
   signal cs     : std_logic := '0';  

   signal oe_s1  : std_logic := '0';
   signal oe_s2  : std_logic := '0';
   signal oe     : std_logic := '0';
   
   signal we_s1  : std_logic := '0';
   signal we_s2  : std_logic := '0';
   signal we     : std_logic := '0';
   
   signal wait_n : std_logic := '1';
   
   --
   -- FSM defines
   --
   type   fsm_state_t is (STATE_IDLE, STATE_READ, STATE_WRITE, STATE_DONE);
   signal fsm_state   : fsm_state_t := STATE_IDLE;
   
   signal wait_cnt    : std_logic_vector(5 downto 0) := (others => '0');  
   
   signal wb_stb      : std_logic := '0';
   signal wb_we       : std_logic := '0';
   signal wb_readdata : std_logic_vector(15 downto 0) := (others => '0');
   
begin

   --
   -- Synchronize CPU control signals
   --
   process (clk_i, rst_i)
   begin
      if (rst_i = '1') then
         cs_s1 <= '0';
         cs_s2 <= '0';
         
         oe_s1 <= '0';
         oe_s2 <= '0';
         
         we_s1 <= '0';
         we_s2 <= '0';
      elsif rising_edge(clk_i) then
         cs_s1 <= not CPU_CS_n;
         cs_s2 <= cs_s1;
         
         oe_s1 <= not CPU_OE_n;
         oe_s2 <= oe_s1;
         
         we_s1 <= not CPU_WE_n;
         we_s2 <= we_s1;
      end if; 
   end process;
   
   cs <= cs_s2; 
   oe <= oe_s2;
   we <= we_s2;
   
   
   --
   -- Single Process FSM
   -- Used to support the following master signals: STB_O, WE_O and ACK_I
   -- 
   process (clk_i, rst_i)     
   begin
      if (rst_i = '1') then
         fsm_state <= STATE_IDLE;
         wb_stb    <= '0';
         wb_we     <= '0';
         wait_n    <= '1';
         
      elsif rising_edge(clk_i) then
                     
         -- Default
         wb_stb <= wb_stb;
         wb_we  <= wb_we;
         wait_n <= '1'; 
         
         case (fsm_state) is
         
            --
            -- STATE_IDLE
            --
            when STATE_IDLE =>
               wb_stb   <= '0';
               wb_we    <= '0';
               wait_cnt <= (others =>'0');
               
               if (cs = '1') then               -- Check for active CPU bus cycle
                  wb_stb <= '1';                -- Signal valid data transfer cycle on the Wishbone bus
                  if (oe = '1') then            -- Check for CPU read access
                     fsm_state <= STATE_READ;
                     wb_we     <= '0';          -- Signal a read cycle on the Wishbone bus
                  end if;
                  if (we = '1') then            -- Check for CPU write access
                     fsm_state <= STATE_WRITE;
                     wb_we     <= '1';          -- Signal a write cycle on the Wishbone bus
                  end if;
               end if;

            --
            -- STATE_READ
            --
            when STATE_READ =>               
               if (wbm_ack_i = '1') then        -- Wait for a normal termination of a Wishbone bus cycle                  
                  fsm_state   <= STATE_DONE;
                  wb_stb      <= '0';           -- Signal end of data transfer on the Wishbone bus
                  wb_readdata <= wbm_dat_i;     -- Wishbone data can be stored
               else               
                  fsm_state <= STATE_READ;
                  wait_n    <= '0';
                  wait_cnt  <= std_logic_vector(unsigned(wait_cnt) + 1);
               end if;

            --
            -- STATE_WRITE
            --
            when STATE_WRITE =>                    
               if (wbm_ack_i = '1') then        -- Wait for a normal termination of a Wishbone bus cycle
                  fsm_state <= STATE_DONE;
                  wb_stb    <= '0';             -- Signal end of data transfer on the Wishbone bus
                  wb_we     <= '0';             -- Not needed, but switch back to read mode
               else
                  fsm_state <= STATE_WRITE;
                  wait_n    <= '0';
                  wait_cnt  <= std_logic_vector(unsigned(wait_cnt) + 1);
               end if;

            --
            -- STATE_DONE
            --
            when STATE_DONE =>
               if ((oe = '0') and (we = '0')) then -- Wait for the end of the CPU bus cycle 
                  fsm_state <= STATE_IDLE;
               else
                  fsm_state <= STATE_DONE;
               end if;
               
            when others =>
               fsm_state <= STATE_IDLE;
               
         end case;
         
         --
         -- Abort the cycle if the bus is occupied too long.
         --
         if (wait_cnt = "111111") then
            fsm_state <= STATE_IDLE;
         end if;
         
      end if;
   end process;
   

   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   -- PERMISSION 3.40
   -- If a MASTER doesnt generate wait states, 
   -- then [STB_O] and [CYC_O] MAY be assigned the same signal.
   wbm_stb_o  <= wb_stb;
   wbm_cyc_o  <= wb_stb;
   
   wbm_we_o   <= wb_we;

   -- Address
   wbm_adr_o  <= CPU_A;
   
   -- Data
   CPU_D      <= wb_readdata when (CPU_CS_n = '0' AND CPU_OE_n = '0') else (others => 'Z');
   wbm_dat_o  <= CPU_D;  

   -- Wait
   CPU_WAIT_n <= wait_n;

end architecture syn;

-- *** EOF ***

