--*****************************************************************************
--*  Copyright (c) 2015 by Michael Fischer (www.emb4fun.de)
--*  All rights reserved.
--*
--*  Redistribution and use in source and binary forms, with or without 
--*  modification, are permitted provided that the following conditions 
--*  are met:
--*  
--*  1. Redistributions of source code must retain the above copyright 
--*     notice, this list of conditions and the following disclaimer.
--*  2. Redistributions in binary form must reproduce the above copyright
--*     notice, this list of conditions and the following disclaimer in the 
--*     documentation and/or other materials provided with the distribution.
--*  3. Neither the name of the author nor the names of its contributors may 
--*     be used to endorse or promote products derived from this software 
--*     without specific prior written permission.
--*
--*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
--*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
--*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
--*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
--*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
--*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
--*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
--*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
--*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
--*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
--*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
--*  SUCH DAMAGE.
--*
--*****************************************************************************
--*  History:
--*
--*  21.07.2015  mifi  First version.
--*****************************************************************************

--*****************************************************************************
--*  DEFINE: Library                                                          *
--*****************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


--*****************************************************************************
--*  DEFINE: Entity                                                           *
--*****************************************************************************

entity wb_core is
   generic (
      ADDRESS_WIDTH : integer := 11
   );
   port (    
      -- Syscon
      clk_i        : in    std_logic := '0';
      rst_i        : in    std_logic := '0';

      -- GPMC bus
      GPMC_A       : in    std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
      GPMC_D       : inout std_logic_vector(15 downto 0);
      GPMC_CS_n    : in    std_logic := '1';
      GPMC_REN_n   : in    std_logic := '1';
      GPMC_WEN_n   : in    std_logic := '1';
      GPMC_RESET_n : in    std_logic := '1';
      GPMC_WAIT_n  : out   std_logic;
      
      gpio0_o      : out   std_logic_vector(15 downto 0);
      gpio0_i      : in    std_logic_vector(15 downto 0) := (others => '0');
      gpio1_o      : out   std_logic_vector(15 downto 0);
      gpio1_i      : in    std_logic_vector(15 downto 0) := (others => '0')
   );
end entity wb_core;


--*****************************************************************************
--*  DEFINE: Architecture                                                     *
--*****************************************************************************

architecture syn of wb_core is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------

   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------

   component cpu_wb_master is
      generic (
         ADDRESS_WIDTH : integer
      );
      port (  
         -- Syscon
         clk_i       : in    std_logic := '0';
         rst_i       : in    std_logic := '0';
      
         -- None Wishbone
         CPU_A       : in    std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
         CPU_D       : inout std_logic_vector(15 downto 0);
         CPU_CS_n    : in    std_logic := '1';
         CPU_OE_n    : in    std_logic := '1';
         CPU_WE_n    : in    std_logic := '1';
         CPU_WAIT_n  : out   std_logic;
      
         -- Wishbone
         wbm_stb_o   : out   std_logic;
         wbm_cyc_o   : out   std_logic;
         wbm_we_o    : out   std_logic;
         wbm_ack_i   : in    std_logic := '0';

         wbm_adr_o   : out   std_logic_vector(ADDRESS_WIDTH-1 downto 0);  

         wbm_dat_o   : out   std_logic_vector(15 downto 0);  
         wbm_dat_i   : in    std_logic_vector(15 downto 0) := (others => '0')
      );
   end component cpu_wb_master;
   
   
   component wb_intercon is
      generic (
         ADDRESS_WIDTH : integer
      );
      port (  
         -- Syscon
         clk_i      : in  std_logic := '0';
         rst_i      : in  std_logic := '0';
      
         -- Wishbone Master
         wbm_stb_i  : in  std_logic := '0';
         wbm_cyc_i  : in  std_logic := '0';
         wbm_we_i   : in  std_logic := '0';
         wbm_ack_o  : out std_logic;
         wbm_adr_i  : in  std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
         wbm_dat_i  : in  std_logic_vector(15 downto 0) := (others => '0');
         wbm_dat_o  : out std_logic_vector(15 downto 0);
         
         -- Wishbone Slave x
         wbs_cyc_o  : out std_logic;
         wbs_we_o   : out std_logic; 
         wbs_dat_o  : out std_logic_vector(15 downto 0);           
      
         -- Wishbone Slave 1
         wbs1_stb_o : out std_logic;
         wbs1_ack_i : in  std_logic := '0';
         wbs1_adr_o : out std_logic_vector(2 downto 0);  
         wbs1_dat_i : in  std_logic_vector(15 downto 0) := (others => '0');
         
         -- Wishbone Slave 2
         wbs2_stb_o : out std_logic;
         wbs2_ack_i : in  std_logic := '0';
         wbs2_adr_o : out std_logic_vector(2 downto 0);  
         wbs2_dat_i : in  std_logic_vector(15 downto 0) := (others => '0')
      );
   end component wb_intercon;


   component wb_gpio_if is
      port (  
         -- Syscon
         clk_i    : in  std_logic := '0';
         rst_i    : in  std_logic := '0';
      
         -- Wishbone
         wb_stb_i : in  std_logic := '0';
         wb_cyc_i : in  std_logic := '0';      -- Not used here
         wb_we_i  : in  std_logic := '0';
         wb_ack_o : out std_logic;
         wb_adr_i : in  std_logic_vector(2 downto 0)  := (others => '0');
         wb_dat_i : in  std_logic_vector(15 downto 0) := (others => '0');
         wb_dat_o : out std_logic_vector(15 downto 0);  
            
         -- None Wishbone
         port_i   : in  std_logic_vector(15 downto 0) := (others => '0');
         port_o   : out std_logic_vector(15 downto 0)
      );
   end component wb_gpio_if;


   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------

   signal wbm_stb        : std_logic := '0';
   signal wbm_cyc        : std_logic := '0';
   signal wbm_we         : std_logic := '0';
   signal wbm_ack        : std_logic := '0';
   signal wbm_adr        : std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
   signal wbm_writedata  : std_logic_vector(15 downto 0) := (others => '0');
   signal wbm_readdata   : std_logic_vector(15 downto 0) := (others => '0');
   
   signal wbs_cyc        : std_logic := '0';
   signal wbs_we         : std_logic := '0';
   signal wbs_writedata  : std_logic_vector(15 downto 0) := (others => '0');
   
   signal gpio0_stb      : std_logic := '0';
   signal gpio0_ack      : std_logic := '0';
   signal gpio0_adr      : std_logic_vector(2 downto 0)  := (others => '0');
   signal gpio0_readdata : std_logic_vector(15 downto 0) := (others => '0');
      
   signal gpio1_stb      : std_logic := '0';
   signal gpio1_ack      : std_logic := '0';
   signal gpio1_adr      : std_logic_vector(2 downto 0)  := (others => '0');
   signal gpio1_readdata : std_logic_vector(15 downto 0) := (others => '0');
      
   signal gpio0_out      : std_logic_vector(15 downto 0) := (others => '0');
   signal gpio1_out      : std_logic_vector(15 downto 0) := (others => '0');
     
begin

   -- Wishbone Master
   inst_master : cpu_wb_master       
      generic map (
         ADDRESS_WIDTH => GPMC_A'length
      )
      port map (  
         -- Syscon
         clk_i       => clk_i,
         rst_i       => rst_i,
         
         -- None Wishbone
         CPU_A       => GPMC_A,
         CPU_D       => GPMC_D,
         CPU_CS_n    => GPMC_CS_n,
         CPU_OE_n    => GPMC_REN_n,
         CPU_WE_n    => GPMC_WEN_n,
         CPU_WAIT_n  => GPMC_WAIT_n,
      
         -- Wishbone Master
         wbm_stb_o   => wbm_stb,
         wbm_cyc_o   => wbm_cyc,
         wbm_we_o    => wbm_we,
         wbm_ack_i   => wbm_ack,
         wbm_adr_o   => wbm_adr,
         wbm_dat_o   => wbm_writedata,
         wbm_dat_i   => wbm_readdata 
      );
      
   -- Wishbone Intercon
   inst_intercon : wb_intercon      
      generic map (
         ADDRESS_WIDTH => GPMC_A'length
      )
      port map (  
         -- Syscon
         clk_i       => clk_i,
         rst_i       => rst_i,
      
         -- Wishbone Master
         wbm_stb_i   => wbm_stb,
         wbm_cyc_i   => wbm_cyc,
         wbm_we_i    => wbm_we,
         wbm_ack_o   => wbm_ack,
         wbm_adr_i   => wbm_adr,
         wbm_dat_i   => wbm_writedata,
         wbm_dat_o   => wbm_readdata,
         
         -- Wishbone Slave x
         wbs_cyc_o   => wbs_cyc,
         wbs_we_o    => wbs_we,
         wbs_dat_o   => wbs_writedata,
      
         -- Wishbone Slave 1
         wbs1_stb_o  => gpio0_stb,
         wbs1_ack_i  => gpio0_ack,
         wbs1_adr_o  => gpio0_adr,
         wbs1_dat_i  => gpio0_readdata,
         
         -- Wishbone Slave 2
         wbs2_stb_o  => gpio1_stb,
         wbs2_ack_i  => gpio1_ack,
         wbs2_adr_o  => gpio1_adr,
         wbs2_dat_i  => gpio1_readdata
      );
      
   -- Wishbone GPIO0
   inst_gpio0 : wb_gpio_if
      port map (  
         -- Syscon
         clk_i    => clk_i,
         rst_i    => rst_i,
      
         -- Wishbone
         wb_stb_i => gpio0_stb,
         wb_cyc_i => wbs_cyc,
         wb_we_i  => wbs_we,
         wb_ack_o => gpio0_ack,
         wb_adr_i => gpio0_adr,
         wb_dat_i => wbs_writedata,
         wb_dat_o => gpio0_readdata,
            
         -- None Wishbone
         port_i   => gpio0_i, 
         port_o   => gpio0_out
      );
      
   -- Wishbone GPIO1
   inst_gpio1 : wb_gpio_if
      port map (  
         -- Syscon
         clk_i    => clk_i,
         rst_i    => rst_i,
      
         -- Wishbone
         wb_stb_i => gpio1_stb,
         wb_cyc_i => wbs_cyc,
         wb_we_i  => wbs_we,
         wb_ack_o => gpio1_ack,
         wb_adr_i => gpio1_adr,
         wb_dat_i => wbs_writedata,
         wb_dat_o => gpio1_readdata,
            
         -- None Wishbone
         port_i   => gpio1_i,
         port_o   => gpio1_out
      );
               
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   gpio0_o <= gpio0_out;
   gpio1_o <= gpio1_out;
         
end architecture syn;

-- *** EOF ***

